@@ -5,7 +5,7 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2023
-;; Version: 0.9
+;; Version: 1.0
 ;; Package-Requires: ((emacs "27.1") (compat "29.1.4.0"))
 ;; Homepage: https://github.com/minad/jinx
 ;; Keywords: convenience, wp
@@ -74,7 +74,9 @@
   (replace-regexp-in-string
    "\\..*\\'" ""
    (or (bound-and-true-p current-locale-environment)
-       (getenv "LANG")
+       (pcase (getenv "LANG")
+         ((or "C" "POSIX") nil)
+         (lang lang))
        "en_US"))
   "Dictionary language codes, as a string separated by whitespace."
   :type 'string)
@@ -88,7 +90,7 @@
                font-lock-string-face)
     (conf-mode font-lock-comment-face
                font-lock-string-face)
-    ;; yaml-mode and yaml-ts-mode are text-modes,
+    ;; `yaml-mode' and `yaml-ts-mode' are text-modes,
     ;; while they should better be conf- or prog-modes.
     (yaml-mode font-lock-comment-face
                font-lock-string-face)
@@ -116,12 +118,10 @@ Set to t to enable camelCase everywhere."
      markdown-plain-url-face markdown-reference-face markdown-url-face)
     (org-mode
      org-block org-block-begin-line org-block-end-line org-code org-cite
-     org-cite-key org-date org-document-info-keyword org-footnote org-formula
-     org-latex-and-related org-link org-macro org-meta-line org-property-value
-     org-ref-cite-face org-special-keyword org-tag org-todo
-     org-todo-keyword-done org-todo-keyword-habt org-todo-keyword-kill
-     org-todo-keyword-outd org-todo-keyword-todo org-todo-keyword-wait
-     org-verbatim
+     org-cite-key org-date org-document-info-keyword org-done org-drawer
+     org-footnote org-formula org-latex-and-related org-link org-macro
+     org-meta-line org-property-value org-special-keyword org-tag org-todo
+     org-verbatim org-warning
      ;; org-modern
      org-modern-tag org-modern-date-active org-modern-date-inactive)
     (tex-mode
@@ -139,7 +139,11 @@ Set to t to enable camelCase everywhere."
      font-lock-constant-face font-lock-warning-face)
     (message-mode
      message-header-cc message-header-name message-header-newsgroups
-     message-header-other message-header-to message-header-xheader))
+     message-header-other message-header-to message-header-xheader
+     message-cited-text-1 message-cited-text-2 message-cited-text-3
+     message-cited-text-4 gnus-cite-1 gnus-cite-2 gnus-cite-3
+     gnus-cite-4 gnus-cite-5 gnus-cite-6 gnus-cite-7 gnus-cite-8
+     gnus-cite-9 gnus-cite-10 gnus-cite-11))
   "Alist of faces per major mode.
 These faces mark regions which should be excluded in spell
 checking."
@@ -148,6 +152,7 @@ checking."
 (defcustom jinx-exclude-regexps
   '((emacs-lisp-mode "Package-Requires:.*$")
     (t "[A-Z]+\\>"         ;; Uppercase words
+       "-+\\>"             ;; Hyphens used as lines or bullet points
        "\\w*?[0-9]\\w*\\>" ;; Words with numbers, hex codes
        "[a-z]+://\\S-+"    ;; URI
        "<?[-+_.~a-zA-Z][-+_.~:a-zA-Z0-9]*@[-.a-zA-Z0-9]+>?" ;; Email
@@ -156,10 +161,9 @@ checking."
   "List of excluded regexps per major mode."
   :type '(alist :key-type symbol :value-type (repeat regexp)))
 
-(defcustom jinx-include-modes
-  '(text-mode prog-mode conf-mode)
-  "List of modes included by `global-jinx-mode'."
-  :type '(repeat symbol))
+(defcustom jinx-suggestion-distance 3
+  "Maximal edit distance for session words to be included in suggestions."
+  :type 'natnum)
 
 (defvar-local jinx-local-words ""
   "File-local words, as a string separated by whitespace.")
@@ -167,6 +171,9 @@ checking."
 ;;;###autoload
 (put 'jinx-local-words 'safe-local-variable #'stringp)
 
+;;;###autoload
+(put 'jinx-mode 'safe-local-variable #'not)
+
 ;;;; Faces
 
 (defgroup jinx-faces nil
@@ -186,7 +193,7 @@ checking."
 
 (defface jinx-highlight
   '((t :inherit isearch))
-  "Face used to highlight current misspelling during correction.")
+  "Face used to highlight current misspelled word during correction.")
 
 (defface jinx-save
   '((t :inherit font-lock-negation-char-face))
@@ -247,6 +254,10 @@ checking."
 
 ;;;; Internal variables
 
+(defvar jinx--reschedule-hooks
+  '(window-state-change-hook window-scroll-functions post-command-hook)
+  "Hooks which reschedule the spell checking timer, see `jinx--reschedule'.")
+
 (defvar jinx--predicates
   (list #'jinx--face-ignored-p
         #'jinx--regexp-ignored-p
@@ -312,8 +323,6 @@ Predicate may return a position to skip forward.")
 (declare-function jinx--mod-describe nil)
 (declare-function jinx--mod-langs nil)
 (declare-function jinx--mod-wordchars nil)
-(declare-function org-fold-core-region "org-fold-core")
-(declare-function org-fold-core-get-regions "org-fold-core")
 
 ;;;; Overlay properties
 
@@ -351,8 +360,15 @@ Predicate may return a position to skip forward.")
 
 (defun jinx--word-valid-p (start)
   "Return non-nil if word at START is valid."
-  (let ((word (buffer-substring-no-properties start (point))))
+  (let ((word (buffer-substring-no-properties start (point)))
+        case-fold-search)
     (or (member word jinx--session-words)
+        ;; Allow capitalized words
+        (and (string-match-p "\\`[[:upper:]][[:lower:]]+\\'" word)
+             (cl-loop
+              for w in jinx--session-words
+              thereis (and (string-equal-ignore-case word w)
+                           (string-match-p "\\`[[:lower:]]+\\'" w))))
         (cl-loop for dict in jinx--dicts
                  thereis (jinx--mod-check dict word)))))
 
@@ -478,8 +494,8 @@ If CHECK is non-nil, always check first."
           (with-delayed-message (1 "Checking...")
             (jinx--check-region start end)))
         (jinx--get-overlays start end visible))
-      (user-error (if visible "No misspelling in visible text"
-                    "No misspelling in whole buffer"))))
+      (user-error (if visible "No misspelled word in visible text"
+                    "No misspelled word in whole buffer"))))
 
 (defun jinx--delete-overlays (start end)
   "Delete overlays between START and END."
@@ -490,8 +506,7 @@ If CHECK is non-nil, always check first."
 (defun jinx--cleanup ()
   "Cleanup all overlays and trigger fontification."
   (with-silent-modifications
-    (save-restriction
-      (widen)
+    (without-restriction
       (jinx--delete-overlays (point-min) (point-max))
       (remove-list-of-text-properties (point-min) (point-max) '(jinx--pending))
       (jinx--in-base-buffer #'jit-lock-refontify))))
@@ -580,6 +595,13 @@ If CHECK is non-nil, always check first."
           (setq mod-file (expand-file-name mod-name))))
       (module-load mod-file))))
 
+(defmacro jinx--correct-guard (&rest body)
+  "Guard BODY during correction loop."
+  `(cl-letf (((symbol-function #'jinx--timer-handler) #'ignore) ;; Inhibit
+             (repeat-mode nil)) ;; No repeating of jinx-next and jinx-previous
+     (unless jinx-mode (jinx-mode 1))
+     ,@body))
+
 (defun jinx--correct-highlight (overlay fun)
   "Highlight and show OVERLAY during FUN."
   (declare (indent 1))
@@ -588,7 +610,10 @@ If CHECK is non-nil, always check first."
     (unwind-protect
         (progn
           (if (and (derived-mode-p #'org-mode)
-                   (fboundp 'org-fold-show-set-visibility))
+                   (fboundp 'org-fold-show-set-visibility)
+                   (fboundp 'org-fold-core-get-regions)
+                   (fboundp 'org-fold-core-region))
+              ;; New Org 9.6 fold-core API
               (let ((regions (delq nil (org-fold-core-get-regions
                                         :with-markers t :from (point-min) :to (point-max)))))
                 (org-fold-show-set-visibility 'canonical)
@@ -616,8 +641,7 @@ If CHECK is non-nil, always check first."
 (defun jinx--recheck-overlays ()
   "Recheck all overlays in buffer after a dictionary update."
   (save-excursion
-    (save-restriction
-      (widen)
+    (without-restriction
       (dolist (ov (overlays-in (point-min) (point-max)))
         (when (eq (overlay-get ov 'category) 'jinx-overlay)
           (goto-char (overlay-end ov))
@@ -634,34 +658,46 @@ If CHECK is non-nil, always check first."
                (not (bound-and-true-p icomplete-mode)))
       (minibuffer-completion-help))))
 
+(defun jinx--add-suggestion (list ht word group)
+  "Add suggestion WORD to LIST and HT.
+The word will be associated with GROUP and get a prefix key."
+  (unless (gethash word ht)
+    (add-text-properties
+     0 (length word)
+     (list 'jinx--group group
+           'jinx--prefix
+           (let ((idx (1+ (hash-table-count ht))))
+             (cond
+              ((< idx 10)
+               (format #("%d " 0 3 (face jinx-key))
+                       idx))
+              ((< (- idx 10) (length jinx--select-keys))
+               (format #("0%c " 0 4 (face jinx-key))
+                       (aref jinx--select-keys (- idx 10)))))))
+     word)
+    (push word list)
+    (puthash word t ht))
+  list)
+
 (defun jinx--correct-suggestions (word)
   "Retrieve suggestions for WORD."
-  (nconc
-   (cl-loop
-    with idx = 1
-    with ht = (make-hash-table :test #'equal)
-    for dict in jinx--dicts
-    for desc = (jinx--mod-describe dict)
-    for group = (format "Suggestions from dictionary ‘%s’ (%s)" (car desc) (cdr desc))
-    nconc
-    (cl-loop
-     for sugg in (jinx--mod-suggest dict word)
-     if (not (gethash sugg ht)) collect
-     (progn
-       (add-text-properties
-        0 (length sugg)
-        (list 'jinx--group group
-              'jinx--prefix
-              (cond ((< idx 10)
-                     (format #("%d " 0 3 (face jinx-key)) idx))
-                    ((< (- idx 10) (length jinx--select-keys))
-                     (format #("0%c " 0 4 (face jinx-key)) (aref jinx--select-keys (- idx 10))))))
-        sugg)
-       (cl-incf idx)
-       (puthash sugg t ht)
-       sugg)))
-   (cl-loop for (key . fun) in jinx--save-keys nconc
-            (ensure-list (funcall fun nil key word)))))
+  (let ((ht (make-hash-table :test #'equal))
+        (list nil))
+    (dolist (dict jinx--dicts)
+      (let* ((desc (jinx--mod-describe dict))
+             (group (format "Suggestions from dictionary ‘%s’ (%s)"
+                            (car desc) (cdr desc))))
+        (dolist (w (jinx--mod-suggest dict word))
+          (setq list (jinx--add-suggestion list ht w group)))))
+    (dolist (w (sort (cl-loop for w in jinx--session-words
+                              for d = (string-distance word w)
+                              if (<= d jinx-suggestion-distance)
+                              collect (cons d w))
+                     #'car-less-than-car))
+      (setq list (jinx--add-suggestion list ht (cdr w) "Suggestions from session")))
+    (nconc (nreverse list)
+           (cl-loop for (key . fun) in jinx--save-keys nconc
+                    (ensure-list (funcall fun nil key word))))))
 
 (defun jinx--correct-affixation (cands)
   "Affixate CANDS during completion."
@@ -682,9 +718,8 @@ If CHECK is non-nil, always check first."
       word
     (get-text-property 0 'jinx--group word)))
 
-(defun jinx--correct-table (word)
-  "Completion table for WORD suggestions."
-  (setq word (jinx--correct-suggestions word))
+(defun jinx--correct-table (suggestions)
+  "Completion table for SUGGESTIONS."
   (lambda (str pred action)
     (if (eq action 'metadata)
         '(metadata (display-sort-function . identity)
@@ -693,31 +728,34 @@ If CHECK is non-nil, always check first."
                    (group-function . jinx--correct-group)
                    (affixation-function . jinx--correct-affixation)
                    (annotation-function . jinx--correct-annotation))
-      (complete-with-action action word str pred))))
+      (complete-with-action action suggestions str pred))))
 
-(defun jinx--correct (overlay info)
+(defun jinx--correct-overlay (overlay &optional info)
   "Correct word at OVERLAY, maybe show prompt INFO."
-  (let* ((word (buffer-substring-no-properties
-                (overlay-start overlay) (overlay-end overlay)))
-         (choice
-          (jinx--correct-highlight overlay
-            (lambda ()
-              (when (or (< (point) (window-start)) (> (point) (window-end nil t)))
-                (recenter))
-              (minibuffer-with-setup-hook
-                  #'jinx--correct-setup
-                (or (completing-read
-                     (format "Correct ‘%s’%s: " word (or info ""))
-                     (jinx--correct-table word)
-                     nil nil nil t word)
-                    word)))))
-         (len (length choice)))
-    (pcase (and (> len 0) (assq (aref choice 0) jinx--save-keys))
-      (`(,key . ,fun)
-       (funcall fun 'save key (if (> len 1) (substring choice 1) word))
-       (jinx--recheck-overlays))
-      ((guard (not (equal choice word)))
-       (jinx--correct-replace overlay choice)))))
+  (catch 'jinx--goto
+    (let* ((word (buffer-substring-no-properties
+                  (overlay-start overlay) (overlay-end overlay)))
+           (choice
+            (jinx--correct-highlight overlay
+              (lambda ()
+                (when (or (< (point) (window-start)) (> (point) (window-end nil t)))
+                  (recenter))
+                (minibuffer-with-setup-hook
+                    #'jinx--correct-setup
+                  (or (completing-read
+                       (format "Correct ‘%s’%s: " word (or info ""))
+                       (jinx--correct-table
+                        (jinx--correct-suggestions word))
+                       nil nil nil t word)
+                      word)))))
+           (len (length choice)))
+      (pcase (and (> len 0) (assq (aref choice 0) jinx--save-keys))
+        (`(,key . ,fun)
+         (funcall fun 'save key (if (> len 1) (substring choice 1) word))
+         (jinx--recheck-overlays))
+        ((guard (not (equal choice word)))
+         (jinx--correct-replace overlay choice)))
+      nil)))
 
 (defun jinx--correct-replace (overlay word)
   "Replace OVERLAY with WORD."
@@ -743,6 +781,16 @@ If CHECK is non-nil, always check first."
   (modify-syntax-entry ?’ "w" jinx--syntax-table)
   (modify-syntax-entry ?. "." jinx--syntax-table))
 
+(defun jinx--bounds-of-word ()
+  "Return bounds of word at point using `jinx--syntax-table'."
+  (save-excursion
+    (save-match-data
+      (with-syntax-table jinx--syntax-table
+        (unless (looking-at-p "\\<")
+          (re-search-backward "\\<"))
+        (when (re-search-forward "\\<\\w+\\>" nil t)
+          (cons (match-beginning 0) (match-end 0)))))))
+
 ;;;; Save functions
 
 (defun jinx--save-action (key word ann)
@@ -797,7 +845,11 @@ If SAVE is non-nil save, otherwise format candidate given action KEY."
 ;;;###autoload
 (defun jinx-languages (langs &optional global)
   "Set languages locally or globally to LANGS.
-With prefix argument GLOBAL change the languages globally."
+LANGS should be one or more language codes as a string, separated
+by whitespace.  When called interactively, the language codes are
+read via `completing-read-multiple'.  If the prefix argument
+GLOBAL is non-nil, the languages are changed globally for all
+buffers.  See also the variable `jinx-languages'."
   (interactive
    (list
     (progn
@@ -810,6 +862,7 @@ With prefix argument GLOBAL change the languages globally."
            (user-error "No languages selected"))
        " "))
     current-prefix-arg))
+  (unless jinx-mode (jinx-mode 1))
   (cond
    (global
     (kill-local-variable 'jinx-languages)
@@ -825,37 +878,66 @@ With prefix argument GLOBAL change the languages globally."
   (jinx--cleanup))
 
 ;;;###autoload
-(defun jinx-correct (&optional all)
-  "Correct nearest misspelled word.
-If prefix argument ALL non-nil correct all misspellings."
+(defun jinx-correct-all ()
+  "Correct all misspelled words in the buffer."
+  (interactive "*")
+  (jinx--correct-guard
+   (let* ((overlays (jinx--force-overlays (point-min) (point-max) :check t))
+          (count (length overlays))
+          (idx 0))
+     (push-mark)
+     (while-let ((ov (nth idx overlays)))
+       (if-let (((overlay-buffer ov))
+                (skip (jinx--correct-overlay ov (format " (%d of %d)" (1+ idx) count))))
+           (setq idx (mod (+ idx skip) count))
+         (cl-incf idx))))))
+
+;;;###autoload
+(defun jinx-correct-nearest ()
+  "Correct nearest misspelled word."
+  (interactive "*")
+  (save-excursion
+    (jinx--correct-guard
+     (let* ((overlays (jinx--force-overlays (window-start) (window-end) :visible t))
+            (count (length overlays))
+            (idx 0))
+       ;; Not using `while-let' is intentional here.
+       (while (when-let ((ov (nth idx overlays)))
+                (if (overlay-buffer ov)
+                    (when-let ((skip (jinx--correct-overlay ov)))
+                      (setq idx (mod (+ idx skip) count)))
+                  (cl-incf idx)))))))) ;; Skip deleted overlay
+
+;;;###autoload
+(defun jinx-correct-word (&optional start end)
+  "Correct word between START and END, by default the word before point.
+Suggest corrections even if the word is not misspelled."
+  (interactive)
+  (unless (and start end)
+    (setf (cons start end) (or (jinx--bounds-of-word)
+                               (user-error "No word at point"))))
+  (save-excursion
+    (jinx--correct-guard
+     (while-let ((skip (jinx--correct-overlay (make-overlay start end))))
+       (forward-to-word skip)
+       (when-let ((bounds (jinx--bounds-of-word)))
+         (setf (cons start end) bounds))))))
+
+;;;###autoload
+(defun jinx-correct (&optional arg)
+  "Correct word depending on prefix ARG.
+This command dispatches to the following commands:
+  - `jinx-correct-nearest': If prefix ARG is nil, correct nearest
+    misspelled word.
+  - `jinx-correct-all': If prefix ARG is 4, corresponding to
+    \\[universal-argument] pressed once, correct all misspelled words.
+  - `jinx-correct-word': If prefix ARG is 16, corresponding to
+    \\[universal-argument] pressed twice, correct word before point."
   (interactive "*P")
-  (unless jinx-mode (jinx-mode 1))
-  (cl-letf* (((symbol-function #'jinx--timer-handler) #'ignore) ;; Inhibit
-             (repeat-mode nil) ;; No repeating of jinx-next and jinx-previous
-             (old-point (point-marker))
-             (overlays
-              (if all
-                  (jinx--force-overlays (point-min) (point-max) :check t)
-                (jinx--force-overlays (window-start) (window-end) :visible t)))
-             (count (length overlays))
-             (idx 0))
-    (when all
-      (push-mark))
-    (unwind-protect
-        (while (when-let ((ov (nth idx overlays)))
-                 (let* ((deleted (not (overlay-buffer ov)))
-                        (skip
-                         (catch 'jinx--goto
-                           (unless deleted
-                             (jinx--correct
-                              ov (and all (format " (%d of %d)" (1+ idx) count)))))))
-                   (cond
-                    ((integerp skip) (setq idx (mod (+ idx skip) count)))
-                    ((or all deleted) (cl-incf idx))))))
-      (if all
-          (jinx--in-base-buffer #'jit-lock-refontify)
-        (goto-char old-point)
-        (jinx--in-base-buffer #'jit-lock-refontify (window-start) (window-end))))))
+  (pcase arg
+    ('nil (jinx-correct-nearest))
+    ('(16) (jinx-correct-word))
+    (_ (jinx-correct-all))))
 
 (defun jinx-correct-select ()
   "Quick selection key for corrections."
@@ -870,10 +952,9 @@ If prefix argument ALL non-nil correct all misspellings."
     (delete-minibuffer-contents)
     (insert word)
     (exit-minibuffer)))
-(put #'jinx-correct-select 'completion-predicate #'ignore)
 
 (defun jinx-next (n)
-  "Go to to Nth next misspelling."
+  "Go to to Nth next misspelled word."
   (interactive "p")
   (unless (= n 0)
     (if (minibufferp)
@@ -884,7 +965,7 @@ If prefix argument ALL non-nil correct all misspellings."
         (goto-char (overlay-end (nth (mod n (length ov)) ov)))))))
 
 (defun jinx-previous (n)
-  "Go to to Nth previous misspelling."
+  "Go to to Nth previous misspelled word."
   (interactive "p")
   (jinx-next (- n)))
 
@@ -910,26 +991,32 @@ If prefix argument ALL non-nil correct all misspellings."
           jinx--include-faces (jinx--mode-list jinx-include-faces)
           jinx--exclude-faces (jinx--mode-list jinx-exclude-faces)
           jinx--camel (or (eq jinx-camel-modes t)
-                          (apply #'derived-mode-p jinx-camel-modes))
+                          (seq-some #'derived-mode-p jinx-camel-modes))
           jinx--session-words (split-string jinx-local-words))
     (jinx--load-dicts)
-    (add-hook 'window-state-change-hook #'jinx--reschedule nil t)
-    (add-hook 'window-scroll-functions #'jinx--reschedule nil t)
-    (add-hook 'post-command-hook #'jinx--reschedule nil t)
+    (dolist (hook jinx--reschedule-hooks)
+      (add-hook hook #'jinx--reschedule nil t))
     (jit-lock-register #'jinx--mark-pending))
    (t
     (mapc #'kill-local-variable '(jinx--exclude-regexp jinx--include-faces
                                   jinx--exclude-faces jinx--camel
                                   jinx--dicts jinx--syntax-table
                                   jinx--session-words))
-    (remove-hook 'window-state-change-hook #'jinx--reschedule t)
-    (remove-hook 'window-scroll-functions #'jinx--reschedule t)
-    (remove-hook 'post-command-hook #'jinx--reschedule t)
+    (dolist (hook jinx--reschedule-hooks)
+      (remove-hook hook #'jinx--reschedule t))
     (jit-lock-unregister #'jinx--mark-pending)
     (jinx--cleanup))))
 
+(defcustom global-jinx-modes '(text-mode prog-mode conf-mode)
+  "List of modes where Jinx should be enabled.
+The variable can either be t, nil or a list of t, nil, mode
+symbols or elements of the form (not modes)."
+  :type '(repeat sexp))
+
 ;;;###autoload
-(define-globalized-minor-mode global-jinx-mode jinx-mode jinx--on :group 'jinx)
+(define-globalized-minor-mode global-jinx-mode
+  jinx-mode jinx--on
+  :group 'jinx)
 
 (defun jinx--on ()
   "Turn `jinx-mode' on."
@@ -937,8 +1024,19 @@ If prefix argument ALL non-nil correct all misspellings."
                       buffer-read-only
                       (buffer-base-buffer) ;; Do not enable in indirect buffers
                       (eq (aref (buffer-name) 0) ?\s)))
-             (apply #'derived-mode-p jinx-include-modes))
+             ;; TODO backport `easy-mmode--globalized-predicate-p'
+             (or (eq t global-jinx-modes)
+                 (eq t (cl-loop for p in global-jinx-modes thereis
+                                (pcase-exhaustive p
+                                  ('t t)
+                                  ('nil 0)
+                                  ((pred symbolp) (and (derived-mode-p p) t))
+                                  (`(not . ,m) (and (seq-some #'derived-mode-p m) 0)))))))
     (jinx-mode 1)))
 
+(put #'jinx-correct-select 'completion-predicate #'ignore)
+(put #'jinx-next 'command-modes '(jinx-mode))
+(put #'jinx-previous 'command-modes '(jinx-mode))
+
 (provide 'jinx)
 ;;; jinx.el ends here
